@@ -24,8 +24,11 @@ SpamAssassin is configured using some traditional UNIX-style configuration
 files, loaded from the /usr/share/spamassassin and /etc/mail/spamassassin
 directories.
 
-The C<#> character starts a comment, which continues until end of line,
-and whitespace in the files is not significant.
+The C<#> character starts a comment, which continues until end of line.
+
+Whitespace in the files is not significant, but please note that starting a
+line with whitespace is deprecated, as we reserve its use for multi-line rule
+definitions, at some point in the future.
 
 Paths can use C<~> to refer to the user's home directory.
 
@@ -39,7 +42,6 @@ Where appropriate, default values are listed in parentheses.
 
 package Mail::SpamAssassin::Conf;
 
-use Carp;
 use strict;
 
 use vars	qw{
@@ -47,7 +49,7 @@ use vars	qw{
 	$type_body_evals $type_full_tests $type_full_evals
 	$type_rawbody_tests $type_rawbody_evals 
 	$type_uri_tests $type_uri_evals
-	$type_rbl_evals $type_rbl_res_evals
+	$type_rbl_evals $type_rbl_res_evals $type_meta_tests
 };
 
 @ISA = qw();
@@ -64,6 +66,7 @@ $type_uri_tests  = 109;
 $type_uri_evals  = 110;
 $type_rbl_evals  = 120;
 $type_rbl_res_evals  = 121;
+$type_meta_tests = 122;
 
 ###########################################################################
 
@@ -72,8 +75,7 @@ sub new {
   $class = ref($class) || $class;
   my $self = { }; bless ($self, $class);
 
-  my $main = shift;     # do not add to class, avoid circular ref
-
+  $self->{errors} = 0;
   $self->{tests} = { };
   $self->{descriptions} = { };
   $self->{test_types} = { };
@@ -92,6 +94,7 @@ sub new {
   $self->{full_evals} = { };
   $self->{rawbody_tests} = { };
   $self->{rawbody_evals} = { };
+  $self->{meta_tests} = { };
 
   # testing stuff
   $self->{regression_tests} = { };
@@ -111,7 +114,7 @@ sub new {
 
   $self->{num_check_received} = 2;
 
-  $self->{razor_config} = $main->sed_path ("~/razor.conf");
+  $self->{razor_config} = undef;
   $self->{razor_timeout} = 10;
   $self->{rbl_timeout} = 30;
 
@@ -121,6 +124,7 @@ sub new {
   $self->{auto_whitelist_factor} = 0.5;
 
   $self->{rewrite_subject} = 1;
+  $self->{detailed_phrase_score} = 0;
   $self->{spam_level_stars} = 1;
   $self->{spam_level_char} = '*';
   $self->{subject_tag} = '*****SPAM*****';
@@ -128,18 +132,25 @@ sub new {
   $self->{use_terse_report} = 0;
   $self->{defang_mime} = 1;
   $self->{skip_rbl_checks} = 0;
+  $self->{dns_available} = "test";
   $self->{check_mx_attempts} = 2;
   $self->{check_mx_delay} = 5;
-  $self->{ok_locales} = '';
+  $self->{ok_locales} = 'all';
   $self->{ok_languages} = '';
   $self->{allow_user_rules} = 0;
   $self->{user_rules_to_compile} = 0;
+  $self->{fold_headers} = 1;
 
   $self->{dcc_body_max} = 999999;
   $self->{dcc_fuz1_max} = 999999;
   $self->{dcc_fuz2_max} = 999999;
   $self->{dcc_add_header} = 0;
   $self->{dcc_timeout} = 10;
+  $self->{dcc_options} = '-R';
+
+  $self->{pyzor_max} = 5;
+  $self->{pyzor_add_header} = 0;
+  $self->{pyzor_timeout} = 10;
 
   $self->{whitelist_from} = { };
   $self->{blacklist_from} = { };
@@ -154,7 +165,8 @@ sub new {
   # this will hold the database connection params
   $self->{user_scores_dsn} = '';
   $self->{user_scores_sql_username} = '';
-  $self->{user_scores_sql_passowrd} = '';
+  $self->{user_scores_sql_password} = '';
+  $self->{user_scores_sql_table} = 'userpref'; # Morgan - default to userpref for backwords compatibility
 
   $self->{_unnamed_counter} = 'aaaaa';
 
@@ -172,17 +184,17 @@ sub mtime {
 ###########################################################################
 
 sub parse_scores_only {
-  my ($self, $rules) = @_;
-  $self->_parse ($rules, 1);
+  my ($self) = @_;
+  $self->_parse ($_[1], 1); # don't copy $rules!
 }
 
 sub parse_rules {
-  my ($self, $rules) = @_;
-  $self->_parse ($rules, 0);
+  my ($self) = @_;
+  $self->_parse ($_[1], 0); # don't copy $rules!
 }
 
 sub _parse {
-  my ($self, $rules, $scoresonly) = @_;
+  my ($self, undef, $scoresonly) = @_; # leave $rules in $_[1]
   local ($_);
 
   my $lang = $ENV{'LC_ALL'};
@@ -194,13 +206,75 @@ sub _parse {
   if ($lang eq 'C') { $lang = 'en_US'; }
   $lang =~ s/[\.\@].*$//;	# .utf8 or @euro
 
-  foreach $_ (split (/\n/, $rules)) {
-    s/\r//g; s/(^|(?<!\\))\#.*$/$1/;
-    s/^\s+//; s/\s+$//; /^$/ and next;
+  my $currentfile = '(no file)';
+  my $skipfile = 0;
+
+  foreach (split (/\n/, $_[1])) {
+    s/(?<!\\)#.*$//; # remove comments
+    s/^\s+|\s+$//g;  # remove leading and trailing spaces (including newlines)
+    next unless($_); # skip empty lines
 
     # handle i18n
-    if (s/^lang\s+(\S\S_\S\S)\s+//) { next if ($lang ne $1); }
-    if (s/^lang\s+(\S\S)\s+//) { my $l = $1; next if ($lang !~ /${l}$/i); }
+    if (s/^lang\s+(\S+)\s+//) { next if ($lang !~ /^$1/i); }
+
+    # Versioning assertions
+    if (/^file\s+start\s+(.+)\s*$/) { $currentfile = $1; next; }
+    if (/^file\s+end/) {
+      $currentfile = '(no file)';
+      $skipfile = 0;
+      next;
+    }
+
+=item require_version n.nn
+
+Indicates that the entire file, from this line on, requires a certain version
+of SpamAssassin to run.  If an older or newer version of SpamAssassin tries to
+read configuration from this file, it will output a warning instead, and
+ignore it.
+
+=cut
+
+    if (/^require[-_]version\s+(.*)\s*$/) {
+      my $req_version = $1 + 0.0;
+      if ($Mail::SpamAssassin::VERSION != $req_version) {
+        warn "configuration file \"$currentfile\" requires version ".
+                "$req_version of SpamAssassin, but this is code version ".
+                "$Mail::SpamAssassin::VERSION. Maybe you need to use ".
+                "the -c switch, or remove the old config files? ".
+                "Skipping this file";
+        $skipfile = 1;
+        $self->{errors}++;
+      }
+      next;
+    }
+
+    if ($skipfile) { next; }
+
+=item version_tag string
+
+This tag is appended to the SA version in the X-Spam-Status header. You should
+include it when modify your ruleset, especially if you plan to distribute it.
+A good choice for I<string> is your last name or your initials followed by a
+number which you increase with each change.
+
+e.g.
+
+  version_tag myrules1    # version=2.41-myrules1
+
+=cut
+
+    if(/^version[-_]tag\s+(.*)$/) {
+      my $tag = lc($1);
+      $tag =~ tr/a-z0-9./_/c;
+      foreach (@Mail::SpamAssassin::EXTRA_VERSION) {
+        if($_ eq $tag) {
+          $tag = undef;
+          last;
+        }
+      }
+      push(@Mail::SpamAssassin::EXTRA_VERSION, $tag) if($tag);
+      next;
+    }
 
     # note: no eval'd code should be loaded before the SECURITY line below.
 ###########################################################################
@@ -219,10 +293,10 @@ Regular expressions are not used for security reasons.
 Multiple addresses per line, separated by spaces, is OK.  Multiple C<whitelist_from> lines is also
 OK.
 
-eg.
-whitelist_from joe@example.com fred@example.com
-whitelist_from simon@example.com
+e.g.
 
+  whitelist_from joe@example.com fred@example.com
+  whitelist_from simon@example.com
 
 =cut
 
@@ -230,15 +304,34 @@ whitelist_from simon@example.com
       $self->add_to_addrlist ('whitelist_from', split (' ', $1)); next;
     }
 
+=item whitelist_from_rcvd lists.sourceforge.net sourceforge.net
+
+Use this to supplement the whitelist_from addresses with a check against the
+Received headers. The first parameter is the address to whitelist, and the
+second is a domain to match in the received headers.
+
+e.g.
+
+  whitelist_from_rcvd joe@example.com  example.com
+  whitelist_from_rcvd axkit.org        sergeant.org
+
+=cut
+
+    if (/^whitelist[-_]from[-_]rcvd\s+(\S+)\s+(\S+)\s*$/) {
+      $self->add_to_addrlist_rcvd ('whitelist_from_rcvd', $1, $2);
+      next;
+    }
+
 =item unwhitelist_from add@ress.com
 
 Used to override a default whitelist_from entry, so for example a distribution whitelist_from
 can be overriden in a local.cf file, or an individual user can override a whitelist_from entry
 in their own .user_prefs file.
 
-eg.
-unwhitelist_from joe@example.com fred@example.com
-unwhitelist_from *@amazon.com
+e.g.
+
+  unwhitelist_from joe@example.com fred@example.com
+  unwhitelist_from *@amazon.com
 
 =cut
 
@@ -263,9 +356,10 @@ Used to override a default blacklist_from entry, so for example a distribution b
 can be overriden in a local.cf file, or an individual user can override a blacklist_from entry
 in their own .user_prefs file.
 
-eg.
-unblacklist_from joe@example.com fred@example.com
-unblacklist_from *@spammer.com
+e.g.
+
+  unblacklist_from joe@example.com fred@example.com
+  unblacklist_from *@spammer.com
 
 =cut
 
@@ -307,7 +401,11 @@ See above.
 =item required_hits n.nn   (default: 5)
 
 Set the number of hits required before a mail is considered spam.  C<n.nn> can
-be an integer or a real number.
+be an integer or a real number.  5.0 is the default setting, and is quite
+aggressive; it would be suitable for a single-user setup, but if you're an ISP
+installing SpamAssassin, you should probably set the default to be something
+much more conservative, like 8.0 or 10.0.  Experience has shown that you
+B<will> get plenty of user complaints otherwise!
 
 =cut
 
@@ -321,12 +419,27 @@ Assign a score to a given test.  Scores can be positive or negative real
 numbers or integers.  C<SYMBOLIC_TEST_NAME> is the symbolic name used by
 SpamAssassin as a handle for that test; for example, 'FROM_ENDS_IN_NUMS'.
 
+Note that test names which begin with '__' are reserved for meta-match
+sub-rules, and are not scored or listed in the 'tests hit' reports.
+
 =cut
 
     if (/^score\s+(\S+)\s+(\-*[\d\.]+)$/) {
       $self->{scores}->{$1} = $2+0.0; next;
     }
 
+=item detailed_phrase_score { 0 | 1 }        (default: 0)
+
+This option displays all matches for "contains phrases frequently found in spam"
+Note that this is disabled by default because it can output huge headers (800
+words and more than 8KB in some cases)
+
+=cut
+
+    if (/^detailed[-_]phrase[-_]score\s+(\d+)$/) {
+      $self->{detailed_phrase_score} = $1+0; next;
+    }
+
 =item rewrite_subject { 0 | 1 }        (default: 1)
 
 By default, the subject lines of suspected spam will be tagged.  This can be
@@ -338,6 +451,18 @@ disabled here.
       $self->{rewrite_subject} = $1+0; next;
     }
 
+=item fold_headers { 0 | 1 }        (default: 1)
+
+By default, the X-Spam-Status header will be whitespace folded, in other words,
+it will be broken up into multiple lines instead of one very long one.
+This can be disabled here.
+
+=cut
+
+   if (/^fold[-_]headers\s+(\d+)$/) {
+     $self->{fold_headers} = $1+0; next;
+   }
+
 =item spam_level_stars { 0 | 1 }        (default: 1)
 
 By default, a header field called "X-Spam-Level" will be added to the message,
@@ -356,14 +481,14 @@ This can be useful for MUA rule creation.
 
 =item spam_level_char { x (some character, unquoted) }        (default: *)
 
-By default, the "X-Spam-Level" header will use a '*' character
-with its length equal to the score of the message.
+By default, the "X-Spam-Level" header will use a '*' character with its length
+equal to the score of the message. Some people don't like escaping *s though, 
+so you can set the character to anything with this option.
+
 In other words, for a message scoring 7.2 points with this option set to .
 
 X-Spam-Level: .......
 
-Some people don't like escaping *'s though, so you can set the character to anything with this option.
-
 =cut
 
    if(/^spam[-_]level[-_]char\s+(.)$/) {
@@ -418,6 +543,25 @@ Content-type header alone, set this to 0.
       $self->{defang_mime} = $1+0; next;
     }
 
+=item dns_available { yes | test[: name1 name2...] | no }   (default: test)
+
+By default, SpamAssassin will query some default hosts on the internet to
+attempt to check if DNS is working on not. The problem is that it can introduce
+some delay if your network connection is down, and in some cases it can wrongly
+guess that DNS is unavailable because the test connections failed.
+SpamAssassin includes a default set of 13 servers, among which 3 are picked
+randomly.
+
+You can however specify your own list by specifying
+
+dns_available test: server1.tld server2.tld server3.tld
+
+=cut
+
+    if (/^dns[-_]available\s+(yes|no|test|test:\s+.+)$/) {
+      $self->{dns_available} = ($1 or "test"); next;
+    }
+
 =item skip_rbl_checks { 0 | 1 }   (default: 0)
 
 By default, SpamAssassin will run RBL checks.  If your ISP already does this
@@ -429,10 +573,10 @@ for you, set this to 1.
       $self->{skip_rbl_checks} = $1+0; next;
     }
 
-=item check_mx_attempts n	(default: 3)
+=item check_mx_attempts n	(default: 2)
 
-By default, SpamAssassin checks the From: address for a valid MX three times,
-waiting 5 seconds each time.
+By default, SpamAssassin checks the From: address for a valid MX this many
+times, waiting 5 seconds each time.
 
 =cut
 
@@ -615,20 +759,17 @@ out
       $self->{rbl_timeout} = $1+0; next;
     }
 
-=item ok_locales xx [ yy zz ... ]		(default: en)
+=item ok_locales xx [ yy zz ... ]		(default: all)
 
 Which locales (country codes) are considered OK to receive mail from.  Mail
 using character sets used by languages in these countries, will not be marked
 as possibly being spam in a foreign language.
 
-SpamAssassin will try to determine the local locale, in order to determine
-which charsets should be allowed by default, but on some OSes it may not be
-able to do this effectively, requiring customisation.
+Note that all ISO-8859-* character sets, and Windows code page character sets,
+are always permitted by default anyway.
 
-All ISO-8859-* character sets, and Windows code page character sets, are
-already permitted by default.
-
-The following locales use additional character sets, and are supported:
+If you wish SpamAssassin to block spam in foreign languages, set this to
+the locale which matches your preference, from the list below:
 
 =over 4
 
@@ -654,10 +795,6 @@ Chinese (both simplified and traditional)
 
 =back
 
-To simply allow all character sets through without giving them points, use
-
-	ok_locales	all
-
 =cut
 
     if (/^ok[-_]locales\s+(.+)$/) {
@@ -686,12 +823,40 @@ C<factor> = 1 means just use the long-term mean; C<factor> = 0 mean just use the
 
 Used to describe a test.  This text is shown to users in the detailed report.
 
+Note that test names which begin with '__' are reserved for meta-match
+sub-rules, and are not scored or listed in the 'tests hit' reports.
+
 =cut
 
     if (/^describe\s+(\S+)\s+(.*)$/) {
       $self->{descriptions}->{$1} = $2; next;
     }
 
+=item tflags SYMBOLIC_TEST_NAME [ { net | nice } ... ]
+
+Used to set flags on a test.  These flags are used in the score-determination back
+end system for details of the test's behaviour.  The following flags can be set:
+
+=over 4
+
+=item  net
+
+The test is a network test, and will not be run in the mass checking system
+or if B<-L> is used, therefore its score should not be modified.
+
+=item  nice
+
+The test is intended to compensate for common false positives, and should be
+assigned a negative score.
+
+=back
+
+=cut
+
+    if (/^tflags\s/) {
+      next;     # ignored in SpamAssassin modules
+    }
+
 =item report ...some text for a report...
 
 Set the report template which is attached to spam mail messages.  See the
@@ -701,9 +866,10 @@ example.
 If you change this, try to keep it under 76 columns (inside the the dots
 below).  Bear in mind that EVERY line will be prefixed with "SPAM: " in order
 to make it clear what's been added, and allow other filters to B<remove>
-spamfilter modifications, so you lose 6 columns right there.  Each C<report>
-line appends to the existing template, so use C<clear-report-template> to
-restart.
+spamfilter modifications, so you lose 6 columns right there. Also note that the
+first line of the report must start with 4 dashes, for the same reason. Each
+C<report> line appends to the existing template, so use
+C<clear-report-template> to restart.
 
 The following template items are supported, and will be filled out by
 SpamAssassin:
@@ -839,6 +1005,59 @@ the results
     }
 
 
+=item pyzor_max NUMBER
+
+Pyzor is a system similar to Razor.  This option sets how often a message's
+body checksum must have been reported to the Pyzor server before SpamAssassin
+will consider the Pyzor check as matched.
+
+The default is 5.
+
+=cut
+
+    if (/^pyzor[-_]max\s+(\d+)/) {
+      $self->{pyzor_max} = $1+0; next;
+    }
+
+=item pyzor_add_header { 0 | 1 }   (default: 0)
+
+Pyzor processing creates a message header containing the statistics for the
+message.  This option sets whether SpamAssassin will add the heading to
+messages it processes.
+
+The default is to not add the header.
+
+=cut
+
+    if (/^pyzor[-_]add[-_]header\s+(\d+)$/) {
+      $self->{pyzor_add_header} = $1+0; next;
+    }
+
+=item pyzor_timeout n              (default: 10)
+
+How many seconds you wait for pyzor to complete before you go on without 
+the results
+
+=cut
+
+    if (/^pyzor[-_]timeout\s*(\d+)\s*$/) {
+      $self->{pyzor_timeout} = $1+0; next;
+    }
+
+
+=item razor_timeout n		(default 10)
+
+How many seconds you wait for razor to complete before you go on without 
+the results
+
+=cut
+
+    if (/^razor[-_]timeout\s*(\d+)\s*$/) {
+      $self->{razor_timeout} = $1; next;
+    }
+
+
+
 =item num_check_received { integer }   (default: 2)
 
 How many received lines from and including the original mail relay
@@ -861,6 +1080,41 @@ See dialup_codes for more details and an example
     #
     if ($scoresonly && !$self->{allow_user_rules}) { goto failed_line; }
 
+=back
+
+=head1 SETTINGS
+
+These settings differ from the ones above, in that they are considered
+'privileged'.  Only users running C<spamassassin> from their procmailrc's or
+forward files, or sysadmins editing a file in C</etc/mail/spamassassin>, can
+use them.   C<spamd> users cannot use them in their C<user_prefs> files, for
+security and efficiency reasons, unless allow_user_rules is enabled (and
+then, they may only add rules from below).
+
+=over 4
+
+=item allow_user_rules { 0 | 1 }		(default: 0)
+
+This setting allows users to create rules (and only rules) in their
+C<user_prefs> files for use with C<spamd>. It defaults to off, because
+this could be a severe security hole. It may be possible for users to
+gain root level access if C<spamd> is run as root. It is NOT a good
+idea, unless you have some other way of ensuring that users' tests are
+safe. Don't use this unless you are certain you know what you are
+doing. Furthermore, this option causes spamassassin to recompile all
+the tests each time it processes a message for a user with a rule in
+his/her C<user_prefs> file, which could have a significant effect on
+server load. It is not recommended.
+
+=cut
+
+
+    if (/^allow[-_]user[-_]rules\s+(\d+)$/) {
+      $self->{allow_user_rules} = $1+0; 
+      dbg("Allowing user rules!"); next;
+    }
+
+
 
 # If you think, this is complex, you should have seen the four previous
 # implementations that I scratched :-)
@@ -875,14 +1129,14 @@ Default:
   "relays.osirusoft.com." => "127.0.0.3" };
 
 WARNING!!! When passing a reference to a hash, you need to put the whole hash in
-one line for the parser to read it correctly (you can check with spamassassin -D
-< mesg)
+one line for the parser to read it correctly (you can check with 
+C<< spamassassin -D < mesg >>)
 
 Set this to what your RBLs return for dialup IPs
 It is used by dialup-firsthop and relay-firsthop rules so that you can match
 DUL codes and compensate DUL checks with a negative score if the IP is a dialup
 IP the mail originated from and it was properly relayed by a hop before reaching
-you (hopefully not your secondary MX :-D)
+you (hopefully not your secondary MX :-)
 The trailing "-firsthop" is magic, it's what triggers the RBL to only be run
 on the originating hop
 The idea is to not penalize (or penalize less) people who properly relayed
@@ -936,41 +1190,6 @@ score Z_FUDGE_DUL_OSIRU_FH	1.5
 
     if ($scoresonly) { dbg("Checking privileged commands in user config"); }
 
-=back
-
-=head1 SETTINGS
-
-These settings differ from the ones above, in that they are considered
-'privileged'.  Only users running C<spamassassin> from their procmailrc's or
-forward files, or sysadmins editing a file in C</etc/mail/spamassassin>, can
-use them.   C<spamd> users cannot use them in their C<user_prefs> files, for
-security and efficiency reasons, unless allow_user_rules is enabled (and
-then, they may only add rules from below).
-
-=over 4
-
-=item allow_user_rules { 0 | 1 }		(default: 0)
-
-This setting allows users to create rules (and only rules) in their
-C<user_prefs> files for use with C<spamd>. It defaults to off, because
-this could be a severe security hole. It may be possible for users to
-gain root level access if C<spamd> is run as root. It is NOT a good
-idea, unless you have some other way of ensuring that users' tests are
-safe. Don't use this unless you are certain you know what you are
-doing. Furthermore, this option causes spamassassin to recompile all
-the tests each time it processes a message for a user with a rule in
-his/her C<user_prefs> file, which could have a significant effect on
-server load. It is not recommended.
-
-=cut
-
-
-    if (/^allow[-_]user[-_]rules\s+(\d+)$/) {
-      $self->{allow_user_rules} = $1+0; 
-      dbg("Allowing user rules!"); next;
-    }
-
-
 
 =item header SYMBOLIC_TEST_NAME header op /pattern/modifiers	[if-unset: STRING]
 
@@ -986,6 +1205,10 @@ C<modifiers> as regexp modifiers in the usual style.
 If the C<[if-unset: STRING]> tag is present, then C<STRING> will
 be used if the header is not found in the mail message.
 
+Note that test names which begin with '__' are reserved for meta-match
+sub-rules, and are not scored or listed in the 'tests hit' reports.
+
+
 =item header SYMBOLIC_TEST_NAME exists:name_of_header
 
 Define a header existence test.  C<name_of_header> is the name of a
@@ -1123,11 +1346,45 @@ Define a full-body eval test.  See above.
       next;
     }
 
+=item meta SYMBOLIC_TEST_NAME boolean expression
+
+Define a boolean expression test in terms of other tests that have
+been hit or not hit.  For example:
+
+meta META1        TEST1 && !(TEST2 || TEST3)
+
+Note that English language operators ("and", "or") will be treated as
+rule names, and that there is no XOR operator.
+
+If you want to define a meta-rule, but do not want its individual sub-rules to
+count towards the final score unless the entire meta-rule matches, give the
+sub-rules names that start with '__' (two underscores).  SpamAssassin will
+ignore these for scoring.
+
+=cut
+
+    if (/^meta\s+(\S+)\s+(.*)$/) {
+      $self->add_test ($1, $2, $type_meta_tests);
+      $self->{user_rules_to_compile} = 1 if $scoresonly;
+      next;
+    }
+
 ###########################################################################
     # SECURITY: allow_user_rules is only in affect until here.
     #
     if ($scoresonly) { goto failed_line; }
 
+=back
+
+=head1 PRIVILEGED SETTINGS
+
+These settings differ from the ones above, in that they are considered 'more
+privileged' -- even more than the ones in the SETTINGS section.  No matter what
+C<allow_user_rules> is set to, these can never be set from a user's
+C<user_prefs> file.
+
+=over 4
+
 
 =item test SYMBOLIC_TEST_NAME (ok|fail) Some string to test against
 
@@ -1154,30 +1411,17 @@ Currently this is the same value Razor itself uses: C<~/razor.conf>.
       $self->{razor_config} = $1; next;
     }
 
-=item razor_timeout n		(default 10)
-
-How many seconds you wait for razor to complete before you go on without 
-the results
-
-=cut
-
-    if (/^razor[-_]timeout\s*(\d+)\s*$/) {
-      $self->{razor_timeout} = $1; next;
-    }
-
 =item dcc_options options
 
 Specify additional options to the dccproc(8) command. Please note that only
 [A-Z -] is allowed (security).
 
-The default is '-R'
+The default is C<-R>
 
 =cut
 
     if (/^dcc_options\s+[A-Z -]+/) {
       $self->{dcc_options} = $1; next;
-    } else {
-      $self->{dcc_options} = '-R';
     }
 
 =item auto_whitelist_path /path/to/file	(default: ~/.spamassassin/auto-whitelist)
@@ -1248,6 +1492,15 @@ The password for the database username, for the above DSN.
       $self->{user_scores_sql_password} = $1; next;
     }
 
+=item user_scores_sql_table tablename
+
+The table user preferences are stored in, for the above DSN.
+
+=cut
+    if(/^user[-_]scores[-_]sql[-_]table\s+(\S+)$/) {
+      $self->{user_scores_sql_table} = $1; next;
+    }
+
 =item spamphrase score phrase ...
 
 A 2-word spam phrase, for the FREQ_SPAM_PHRASE test.
@@ -1269,7 +1522,15 @@ The highest score of any of the spamphrases.  Used for scaling.
 ###########################################################################
 
 failed_line:
-    dbg ("Failed to parse line in SpamAssassin configuration, skipping: $_");
+    my $msg = "Failed to parse line in SpamAssassin configuration, ".
+                        "skipping: $_";
+
+    if ($self->{lint_rules}) {
+      warn $msg."\n";
+    } else {
+      dbg ($msg);
+    }
+    $self->{errors}++;
   }
 }
 
@@ -1325,8 +1586,10 @@ sub finish_parsing {
     elsif ($type == $type_full_evals) { $self->{full_evals}->{$name} = $text; }
     elsif ($type == $type_uri_tests)  { $self->{uri_tests}->{$name} = $text; }
     # elsif ($type == $type_uri_evals)  { $self->{uri_evals}->{$name} = $text; }
+    elsif ($type == $type_meta_tests) { $self->{meta_tests}->{$name} = $text; }
     else {
       # 70 == SA_SOFTWARE
+      $self->{errors}++;
       sa_die (70, "unknown type $type for $name: $text");
     }
   }
@@ -1346,6 +1609,17 @@ sub add_to_addrlist {
   }
 }
 
+sub add_to_addrlist_rcvd {
+  my ($self, $listname, $addr, $domain) = @_;
+  
+  my $re = lc $addr;
+  $re =~ s/[\000\\\(]/_/gs;			# paranoia
+  $re =~ s/([^\*_a-zA-Z0-9])/\\$1/g;		# escape any possible metachars
+  $re =~ s/\*/\.\*/g;				# "*" -> "any string"
+  $self->{$listname}->{$addr}{re} = qr/^${re}$/;
+  $self->{$listname}->{$addr}{domain} = $domain;
+}
+
 sub remove_from_addrlist {
   my ($self, $singlelist, @addrs) = @_;
   
